#!/www/python/bin/python
"""$URL: svn+ssh://svn/repos/trunk/durus/file_storage.py $
$Id: file_storage.py 25046 2004-09-08 18:46:23Z nascheme $
"""
import os
import tempfile
from sets import Set
from durus.connection import ROOT_OID
from durus.serialize import split_oids, unpack_record
from durus.storage import Storage
from durus.utils import p32, u32, p64, u64

def readn(fp, n):
    """Read 'n' bytes from a file.  Raises IOError if fewer bytes than 'n'
    bytes are available."""
    s = fp.read(n)
    if len(s) != n:
        raise IOError, 'short read'
    return s

MAGIC = "DFS10\0"

if os.name == 'posix':
    import fcntl
    def lock_file(fp):
        fcntl.flock(fp, fcntl.LOCK_EX | fcntl.LOCK_NB)
    def unlock_file(fp):
        pass
    RENAME_OPEN_FILE = True
elif os.name == 'nt':
    import win32con, win32file, pywintypes # http://sf.net/projects/pywin32/
    def lock_file(fp):
        win32file.LockFileEx(win32file._get_osfhandle(fp.fileno()),
                             (win32con.LOCKFILE_EXCLUSIVE_LOCK |
                              win32con.LOCKFILE_FAIL_IMMEDIATELY),
                             0, -65536, pywintypes.OVERLAPPED())
    def unlock_file(fp):
        win32file.UnlockFileEx(win32file._get_osfhandle(fp.fileno()),
                               0, -65536, pywintypes.OVERLAPPED())
    RENAME_OPEN_FILE = False
else:
    def lock_file(fp):
        raise RuntimeError("Sorry, don't know how to lock files on your OS")
    def unlock_file(fp):
        pass
    RENAME_OPEN_FILE = False

if hasattr(os, 'fsync'):
    fsync = os.fsync
else:
    def fsync(fd):
        pass

class FileStorage(Storage):
    """

    The file contains a sequence of transaction records.
    Each transaction record is a sequence of object records
    followed by a 4-byte zero terminator.
    Object records have the following structure:
      1) The number of bytes in the next four components of the object
         record (4 bytes, unsigned, big-endian).
      2) The transaction identifier (8 bytes, unsigned, big-endian).
         This will be the same for every object record in the
         transaction record.
      3) The object identifier (8 bytes, unsigned, big-endian).
      4) The pickled object state.  This is normally a the pickle of
         the object's __dict__, pickled using protocol 2, with a
         customized pickler that stores oid and class for each
         reference to a persistent object.
      5) A sequence of object identifiers (each 8 bytes, unsigned, big-endian)
         of objects with references in the pickled object state.
         These referenced object identifiers can be collected directly
         while unpickling pickled object state, but they are included
         directly here for faster access.

    Instance attributes:
      fp : file
      index : { oid:string : offset:int }
        Gives the offset of the current version of each oid.
      pending_records : [str]
        Object records are accumulated here during a commit.
    """

    def __init__(self, filename, readonly=False, repair=False):
        """(filename:str, readonly:bool=False, repair:bool=False)
        """
        self.tid = 0
        self.oid = 0
        self.filename = filename
        if readonly:
            self.fp = open(self.filename, 'r')
        else:
            self.fp = open(self.filename, 'a+')
            try:
                lock_file(self.fp)
            except IOError:
                self.fp.close()
                raise RuntimeError(
                    "\n  %s is locked."
                    "\n  There is probably a Durus storage server (or a client)"
                    "\n  using it.\n" % self.filename)
            if os.fstat(self.fp.fileno()).st_size == 0:
                self.fp.write(MAGIC)
        self.pending_records = []
        self.repair = repair
        self._build_index()
        #print '  len(index)==%s' % len(self.index)

    def _build_index(self):
        self.index = {}
        self.fp.seek(0)
        if self.fp.read(len(MAGIC)) != MAGIC:
            raise IOError, "invalid storage (missing magic)"
        max_tid = 0
        while 1:
            # Read one transaction each time here.
            transaction_offset = self.fp.tell()
            current_tid = None
            oids = {}
            try:
                while 1:
                    object_record_offset = self.fp.tell()
                    size = u32(readn(self.fp, 4))
                    if size == 0:
                        break # normal termination
                    if 0 < size < 16:
                        raise ValueError("Bad record size")
                    trecord = readn(self.fp, size)
                    tid = trecord[0:8]
                    oid = trecord[8:16]
                    if current_tid is None:
                        current_tid = tid
                    else:
                        if current_tid != tid:
                            raise ValueError("Bad tid")
                    if oid in oids:
                        raise ValueError("Object duplicated")
                    oids[oid] = object_record_offset
                # We've reached the normal end of a transaction.
                if current_tid is not None:
                    # The file was not empty.
                    self.index.update(oids)
                    max_tid = max(max_tid, u64(current_tid))
                    oids.clear()
            except (ValueError, IOError), exc:
                if self.fp.tell() > transaction_offset:
                    if not self.repair:
                        raise
                    # The transaction was malformed. Attempt repair.
                    if self.fp.mode == 'r':
                        raise RuntimeError(
                            "Can't repair readonly file.\n%s" % exc)
                    self.fp.seek(transaction_offset)
                    self.fp.truncate()
                break
        self.tid = max_tid
        max_oid = 0
        for oid in self.index:
            max_oid = max(max_oid, u64(oid))
        self.oid = max_oid

    def _get_tid(self):
        return p64(self.tid)

    def get_size(self):
        return len(self.index)

    def new_oid(self):
        self.oid += 1
        return p64(self.oid)

    def load(self, oid):
        if self.fp is None:
            raise IOError, 'storage is closed'
        offset = self.index[oid]
        self.fp.seek(offset)
        size = u32(self.fp.read(4))
        return self.fp.read(size)

    def begin(self):
        """Begin a commit."""
        self.tid += 1

    def store(self, record):
        """Add a record during a commit."""
        self.pending_records.append(record)

    def end(self, handle_invalidations=None):
        """Complete a commit.
        A FileStorage is the storage of one StorageServer or one
        Connection, so there can never be any invalidations to handle.
        """
        if self.fp is None:
            raise IOError, 'storage is closed'
        self.fp.seek(0, 2)
        tid, index = self._write_transaction(self.fp, self.pending_records)
        self.fp.flush()
        fsync(self.fp)
        self.index.update(index)
        del self.pending_records[:]
        return tid

    def _write_transaction(self, file, records):
        tid = self._get_tid()
        index = {}
        for record in records:
            oid = record[:8]
            index[oid] = file.tell()
            trecord = tid + record
            file.write(p32(len(trecord)))
            file.write(trecord)
        file.write(p32(0)) # terminator
        return tid, index

    def sync(self):
        """() -> tid:str, invalidations:[str]
        A FileStorage is the storage of one StorageServer or one
        Connection, so there can never be any invalidations to transfer.
        """
        return self._get_tid(), []

    def _get_pack_names(self):
        """() -> prepack_name:str, pack_name:str
        Return the names of files used during a pack.
        The `prepack_name` is for a copy of the file as it was before
        packing.  The `pack_name` is used to hold the newly packed file.
        """
        return self.filename + '.prepack', self.filename + '.pack'

    def pack(self):
        """Perform a pack on the storage.
        All current object records reachable from the root object
        are moved into one big transaction record.
        """
        assert not self.pending_records
        if self.fp is None:
            raise IOError, 'storage is closed'
        if self.fp.mode == 'r':
            raise IOError, "read-only storage"
        prepack_name, pack_name = self._get_pack_names()
        packed = open(pack_name, 'w+')
        lock_file(packed)
        packed.write(MAGIC)
        def generate_records():
            todo = [ROOT_OID]
            seen = Set()
            while todo:
                oid = todo.pop()
                if oid in seen:
                    continue
                seen.add(oid)
                record = self.load(oid)[8:]
                record_oid, data, refdata = unpack_record(record)
                assert oid == record_oid
                todo.extend(split_oids(refdata))
                yield record
        tid, index = self._write_transaction(packed, generate_records())
        packed.flush()
        fsync(packed)
        if not RENAME_OPEN_FILE:
            unlock_file(packed)
            packed.close()
        unlock_file(self.fp)
        self.fp.close()
        os.rename(self.filename, prepack_name)
        os.rename(pack_name, self.filename)
        if RENAME_OPEN_FILE:
            self.fp = packed
        else:
            self.fp = open(self.filename, 'a+')
            lock_file(self.fp)
        self.index = index

    def gen_oid_record(self):
        """() -> sequence([(oid:str, record:str)])
        Generate oid, record pairs, for all oids in the database.
        Note that this may include oids that are not reachable from
        the root object.
        """
        for oid in self.index:
            yield oid, self.load(oid)

    def close(self):
        if self.fp is not None:
            unlock_file(self.fp)
            self.fp.close()
            self.fp = None


class TempFileStorage(FileStorage):

    def __init__(self):
        (fd, name) = tempfile.mkstemp(suffix='.tmp.durus')
        FileStorage.__init__(self, name)
        os.close(fd)

    _unlink = os.unlink

    def close(self):
        if self.fp is not None:
            FileStorage.close(self)
            self._unlink(self.filename)

    def __del__(self):
        self.close()

    def pack(self):
        FileStorage.pack(self)
        for name in self._get_pack_names():
            if os.path.exists(name):
                os.unlink(name)





#!/www/python/bin/python
"""
$URL: svn+ssh://svn/repos/trunk/durus/test/test_utils.py $
$Id: test_utils.py 23882 2004-04-05 20:47:59Z dbinger $
"""
from sancho.unittest import TestScenario, parse_args, run_scenarios
from durus.utils import format_oid, u64, p64, u32, p32

tested_modules = ["durus.utils"]

class Test (TestScenario):

    def check_format_oid(self):
        self.test_val("format_oid('A'*8)", '4702111234474983745')

    def check_p64_u64(self):
        for x in range(3):
            self.test_val('len(p64(x))', 8)
            self.test_val('u64(p64(x))', x)

    def check_p32_u32(self):
        for x in range(3):
            assert len(p32(x)) == 4
            assert x == u32(p32(x))


if __name__ == "__main__":
    (scenarios, options) = parse_args()
    run_scenarios(scenarios, options)




#!/www/python/bin/python
"""
$URL: svn+ssh://svn/repos/trunk/durus/test/utest_btree.py $
$Id: utest_btree.py 25044 2004-09-08 14:36:06Z dbinger $
"""

from random import randint
from dulcinea.utest import UTest
from durus.btree import BTree, BNode
from durus.connection import Connection
from durus.file_storage import TempFileStorage

class CoverageTest(UTest):

    def delete_case_1(self):
        bt = BTree()
        bt[1] = 2
        del bt[1]

    def delete_keyerror(self):
        bt = BTree()
        try:
            del bt[1]
        except KeyError, e:
            assert str(e) == '1'

    def delete_case_2a(self):
        bt = BTree(BNode)
        map(bt.add, 'jklmoab')
        del bt['k']

    def delete_case_2b(self):
        bt = BTree(BNode)
        map(bt.add, 'abcdef')
        assert bt.root.items == [('b', True), ('d', True)]
        del bt['d']

    def delete_case_2c(self):
        bt = BTree(BNode)
        map(bt.add, 'abcdefghi')
        assert bt.root.items == [('d', True)]
        del bt['d']

    def _delete_case_3(self):
        bt = BTree(BNode)
        map(bt.add, range(100))
        assert bt.root.items == [(31, True), (63, True)]
        assert [n.items for n in bt.root.nodes] == [
            [(15, True)], [(47, True)], [(79, True)]]
        assert [[n.items for n in node.nodes]
                for node in bt.root.nodes] == [
            [[(7, True)], [(23, True)]],
            [[(39, True)], [(55, True)]],
            [[(71, True)], [(87, True)]]]
        return bt

    def delete_case_3a1(self):
        bt = self._delete_case_3()
        del bt[39]
        del bt[55]

    def delete_case_3a2(self):
        bt = self._delete_case_3()
        del bt[39]
        del bt[7]

    def delete_case_3b1(self):
        bt = self._delete_case_3()
        del bt[39]

    def delete_case_3b2(self):
        bt = self._delete_case_3()
        del bt[7]

    def find_extremes(self):
        bt = BTree()
        map(bt.add, range(100))
        assert bt.get_min_item() == (0, True)
        assert bt.get_max_item() == (99, True)

    def iter(self):
        bt = BTree()
        map(bt.add, range(100))
        assert list(bt) == list(bt.iterkeys())
        assert list(bt.iteritems()) == zip(bt, bt.itervalues())
        assert list(bt.iterkeys()) == bt.keys()
        assert list(bt.itervalues()) == bt.values()
        assert list(bt.iteritems()) == bt.items()

    def search(self):
        bt = BTree(BNode)
        map(bt.add, range(100))
        assert bt[1] == True
        try:
            assert bt[-1]
        except KeyError:
            pass

    def insert_again(self):
        bt = BTree(BNode)
        bt[1] = 2
        bt[1] = 3
        assert bt[1] == 3
        assert list(bt) == [1], list(bt)

    def get(self):
        bt = BTree()
        map(bt.add, range(10))
        assert bt.get(2) == True
        assert bt.get(-1) == None
        assert bt.get(-1, 5) == 5

    def contains(self):
        bt = BTree()
        map(bt.add, range(10))
        assert 2 in bt
        assert -1 not in bt


class SlowTest(UTest):

    def slow(self):
        bt = BTree()
        print 'bt = BTree()'
        d = {}
        number = 0
        limit = 100000
        for k in xrange(limit*10):
            number = randint(0, limit)
            if number in bt:
                assert number in d
                if randint(0, 1) == 1:
                    del bt[number]
                    del d[number]
                    print 'del bt[%s]' % number
            else:
                if number in d:
                    print number
                    print number in bt
                    print number in d
                    assert number not in d
                bt[number] = 1
                d[number] = 1
                print 'bt[%s] = 1' % number
            if k % limit == 0:
                d_items = d.items()
                d_items.sort()
                assert d_items == bt.items()

class DurusTest(UTest):

    def _pre(self):
        self.connection = Connection(TempFileStorage())

    def _post(self):
        del self.connection

    def a(self):
        bt = self.connection.get_root()['bt'] = BTree()
        t = bt.root.minimum_degree
        assert self.connection.get_cache_count() == 1
        for x in range(2 * t - 1):
            bt.add(x)
        self.connection.commit()
        assert self.connection.get_cache_count() == 3
        bt.add(2 * t - 1)
        self.connection.commit()
        assert self.connection.get_cache_count() == 5


if __name__ == '__main__':
    CoverageTest()
    DurusTest()
    for k in xrange(1000):
        print k,
        SlowTest()



#!/www/python/bin/python
"""
$URL: svn+ssh://svn/repos/trunk/durus/test/test_connection.py $
$Id: test_connection.py 23973 2004-04-13 16:44:27Z nascheme $
"""

from sancho.unittest import TestScenario, parse_args, run_scenarios
from durus.file_storage import TempFileStorage
from durus.utils import p64
from durus.connection import Connection
from durus.persistent import Persistent

tested_modules = ["durus.connection"]

class Test (TestScenario):

    def check_connection(self):
        storage = TempFileStorage()
        self.test_stmt("self.conn=conn=Connection(storage)")
        self.test_stmt("self.root=root=conn.get_root()")
        self.test_val("root._p_is_ghost()", True)
        self.test_bool("root is conn.get(p64(0))")
        self.test_bool("root is conn.get(0)")
        self.test_bool("conn is root._p_connection")
        self.test_val("conn.get(p64(1))", None)
        self.test_stmt("conn.abort()")
        self.test_stmt("conn.commit()")
        self.test_val("root._p_is_ghost()", True)
        self.test_stmt("root['a'] = Persistent()")
        self.test_val("root._p_is_unsaved()", True)
        self.test_val("root['a']._p_is_unsaved()", True)
        self.test_stmt("root['a'].f=2")
        self.test_true("conn.changed.values() == [root]")
        self.test_stmt("conn.commit()")
        self.test_true("root._p_is_saved()")
        self.test_true("conn.changed.values() == []")
        self.test_stmt("root['a'] = Persistent()")
        self.test_true("conn.changed.values() == [root]")
        self.test_stmt("root['b'] = Persistent()")
        self.test_stmt("root['a'].a = 'a'")
        self.test_stmt("root['b'].b = 'b'")
        self.test_stmt("conn.commit()")
        self.test_stmt("root['a'].a = 'a'")
        self.test_stmt("root['b'].b = 'b'")
        self.test_stmt("conn.abort()")
        self.test_stmt("conn.shrink_cache()")
        self.test_stmt("root['b'].b = 'b'")
        self.test_stmt("del conn")

    def check_shrink(self):
        storage = TempFileStorage()
        self.test_stmt("self.conn=conn=Connection(storage, cache_size=3)")
        self.test_stmt("self.root=root=conn.get_root()")
        self.test_stmt("root['a'] = Persistent()")
        self.test_stmt("root['b'] = Persistent()")
        self.test_stmt("root['c'] = Persistent()")
        assert self.root._p_is_unsaved()
        self.test_stmt("conn.commit()")
        self.test_stmt("root['a'].a = 1")
        self.test_stmt("conn.commit()")
        self.test_stmt("root['b'].b = 1")
        self.test_stmt("root['c'].c = 1")
        self.test_stmt("root['d'] = Persistent()")
        self.test_stmt("root['e'] = Persistent()")
        self.test_stmt("root['f'] = Persistent()")
        self.test_stmt("conn.commit()")
        self.test_stmt("root['f'].f = 1")
        self.test_stmt("root['g'] = Persistent()")
        self.test_stmt("conn.commit()")
        self.test_stmt("conn.pack()")

if __name__ == "__main__":
    (scenarios, options) = parse_args()
    options.show_coverage = options.show_coverage_lines = 1
    run_scenarios(scenarios, options)




#!/www/python/bin/python
"""
$URL: svn+ssh://svn/repos/trunk/durus/run_durus.py $
$Id: run_durus.py 25008 2004-08-27 22:20:28Z dbinger $
"""
import sys
import socket
from optparse import OptionParser
from durus.storage_server import DEFAULT_PORT, DEFAULT_HOST, StorageServer
from durus.file_storage import FileStorage, TempFileStorage
from durus.logger import log, logger, direct_output

def start_durus(logfile, logginglevel, file, repair, readonly, host, port):
    if logfile is None:
        logfile = sys.stderr
    else:
        logfile = open(logfile, 'a+')
    direct_output(logfile)
    logger.setLevel(logginglevel)
    if file is None:
        storage = TempFileStorage()
    else:
        storage = FileStorage(file, repair=repair,
                              readonly=readonly)
    log(20, 'Storage file=%s host=%s port=%s', storage.fp.name, host, port)
    StorageServer(storage, host=host, port=port).serve()

def stop_durus(host, port):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        sock.connect((host, port))
    except socket.error, e:
        raise SystemExit("Durus server %s:%s doesn't seem to be running." %
                         (host, port))
    sock.send('Q') # graceful exit message
    sock.close()

def run_durus_main():
    parser = OptionParser()
    parser.set_description('Run a Durus Server')
    parser.add_option(
        '--port', dest='port', default=DEFAULT_PORT, type='int',
        help='Port to listen on. (default=%s)' % DEFAULT_PORT)
    parser.add_option(
        '--file', dest='file', default=None,
        help='If this is not given, the storage is in a new temporary file.')
    parser.add_option(
        '--host', dest='host', default=DEFAULT_HOST,
        help='Host to listen on. (default=%s)' % DEFAULT_HOST)
    logginglevel = logger.getEffectiveLevel()
    parser.add_option(
        '--logginglevel', dest='logginglevel', default=logginglevel, type='int',
        help=('Logging level. Lower positive numbers log more. (default=%s)' %
              logginglevel))
    parser.add_option(
        '--logfile', dest='logfile', default=None,
        help=('Log file. (default=stderr)'))
    parser.add_option(
        '--repair', dest='repair', action='store_true',
        help=('Repair the filestorage by truncating to remove anything '
              'that is malformed.  Without this option, errors '
              'will cause the program to report and terminate without '
              'attempting any repair.'))
    parser.add_option(
        '--readonly', dest='readonly', action='store_true',
        help='Open the file in read-only mode.')
    parser.add_option(
        '--stop', dest='stop', action='store_true',
        help='Instead of starting the server, try to stop a running one.')
    (options, args) = parser.parse_args()
    if not options.stop:
        start_durus(options.logfile,
                    options.logginglevel,
                    options.file,
                    options.repair,
                    options.readonly,
                    options.host,
                    options.port)
    else:
        stop_durus(options.host,
                   options.port)


if __name__ == '__main__':
    run_durus_main()




#!/www/python/bin/python
"""
$URL: svn+ssh://svn/repos/trunk/durus/test/test_storage_server.py $
$Id: test_storage_server.py 23897 2004-04-06 20:56:28Z dbinger $
"""
from random import choice
from sancho.unittest import TestScenario, parse_args, run_scenarios
from durus.storage_server import StorageServer, recv
from durus.file_storage import TempFileStorage

tested_modules = ["durus.storage_server"]

class Test (TestScenario):

    def check_storage_server(self):
        storage = TempFileStorage()
        host = '127.0.0.1'
        port = 2972
        self.test_stmt("server=StorageServer(storage, host=host, port=port)")

    def check_receive(self):
        class Dribble:
            def recv(x, n):
                return choice(['a', 'bb'])[:n]
        fake_socket = Dribble()
        self.test_stmt("recv(fake_socket, 30)")


if __name__ == "__main__":
    (scenarios, options) = parse_args()
#    options.show_coverage = options.show_coverage_lines = 1
    run_scenarios(scenarios, options)




#!/www/python/bin/python
"""
$URL: svn+ssh://svn/repos/trunk/durus/test/test_client_storage.py $
$Id: test_client_storage.py 24943 2004-08-18 12:12:24Z dbinger $
"""
from time import sleep
from os import kill
from popen2 import Popen4
from signal import SIGTERM
from sancho.unittest import TestScenario, parse_args, run_scenarios
from durus import run_durus
from durus.client_storage import ClientStorage
from durus.serialize import pack_record
from durus.utils import p64

tested_modules = ["durus.client_storage"]

class Test (TestScenario):

    def setup(self):
        self.port = 9123
        self.server = Popen4('python %s --port=%s' % (
            run_durus.__file__, self.port))
        sleep(3) # wait for bind

    def shutdown(self):
        run_durus.stop_durus("", self.port)

    def check_client_storage(self):
        self.test_stmt("b=ClientStorage(port=%s)" % self.port)
        self.test_val("b.new_oid()", p64(1))
        self.test_val("b.new_oid()", p64(2))
        self.test_exc("b.load(p64(0))", KeyError)
        record = pack_record(p64(0), 'ok', '')
        self.test_stmt("b.begin()")
        self.test_stmt("b.store(record)")
        self.test_stmt("b.end()")
        self.test_stmt("b.load(p64(0))")
        self.test_stmt("b.sync()")
        self.test_stmt("b.begin()")
        self.test_stmt("b.store(pack_record(p64(1), 'no', ''))")
        self.test_stmt("b.end()")
        self.test_stmt("b.pack()")

if __name__ == "__main__":
    (scenarios, options) = parse_args()
    options.show_coverage = options.show_coverage_lines = 1
    run_scenarios(scenarios, options)




"""$URL: svn+ssh://svn/repos/trunk/durus/error.py $
$Id: error.py 24794 2004-07-26 15:15:23Z nascheme $
"""

from durus.utils import format_oid

class DurusError(StandardError):
    """Durus error."""


class DurusKeyError(KeyError, DurusError):
    """Key not found in database."""

    def __str__(self):
        return format_oid(self.args[0])

class InvalidObjectReference(DurusError):
    """
    An object contains an invalid reference to another object.

    A reference is invalid if it refers to an object managed
    by a different database connection.

    Instance attributes:
      obj: Persistent
        is the object for which the reference is invalid.
      connection: Connection
        the connection that attempted to store it.
    
    obj._p_connection != connection
    """

    def __init__(self, obj, connection):
        self.obj = obj
        self.connection = connection

    def __str__(self):
        return "Invalid reference to %r with a connection %r." % (
            self.obj,
            self.obj._p_connection)


class ConflictError(DurusError):
    """
    Two transactions tried to modify the same object at once.
    This transaction should be resubmitted.
    The object passed to the constructor should be an instance of Persistent.
    """
    def __init__(self, oids):
        self.oids = oids

    def __str__(self):
        if len(self.oids) > 1:
            s = "oids=[%s ...]"
        else:
            s = "oids=[%s]"
        return s % format_oid(self.oids[0])


class ReadConflictError(ConflictError):
    """
    Conflict detected when object was loaded.
    An attempt was made to read an object that has changed in another
    transaction (eg. another process).
    """


class ProtocolError(DurusError):
    """
    An error occurred during communication between the storage server
    and the client.
    """



#!/www/python/bin/python
"""
$URL: svn+ssh://svn/repos/trunk/durus/test/test_serialize.py $
$Id: test_serialize.py 23910 2004-04-07 13:39:44Z dbinger $
"""
from sancho.unittest import TestScenario, parse_args, run_scenarios
from durus.connection import ROOT_OID
from durus.error import InvalidObjectReference
from durus.persistent import Persistent
from durus.serialize import findrefs, ObjectWriter, ObjectReader, pack_record, \
     unpack_record, split_oids

tested_modules = ["durus.serialize"]

class Test (TestScenario):

    def check_object_writer(self):
        class FakeConnection:
            def new_oid(self):
                return ROOT_OID
        connection = FakeConnection()
        self.test_stmt("self.s=s=ObjectWriter(connection)")
        x = Persistent()
        self.test_val("x._p_connection", None)
        x._p_oid = ROOT_OID
        x._p_connection = connection
        self.test_val("s._persistent_id(x)", (ROOT_OID, Persistent))
        x._p_connection = ROOT_OID
        # connection of x no longer matches connection of s.
        self.test_exc("s._persistent_id(x)", InvalidObjectReference)
        x.a = Persistent()
        self.test_val("s.get_state(x)",
                      ('\x80\x02cdurus.persistent\nPersistent\nq\x01.\x80\x02}q\x02U'
                       '\x01aU\x08\x00\x00\x00\x00\x00\x00\x00\x00q\x03h\x01\x86Qs.',
                       '\x00\x00\x00\x00\x00\x00\x00\x00'))
        self.test_val("list(s.gen_new_objects(x))", [x, x.a])
        # gen_new_objects() can only be called once.
        self.test_exc("s.gen_new_objects(3)", RuntimeError)
        self.test_stmt("s.close()")

    def check_object_reader(self):
        class FakeConnection:
            pass
        self.test_stmt("self.r = r = ObjectReader(FakeConnection())")
        root = '\x80\x02cdurus.persistent_dict\nPersistentDict\nq\x01.\x80\x02}q\x02U\x04dataq\x03}q\x04s.\x00\x00\x00\x00'
        self.test_true("r.get_ghost(root)._p_is_ghost()")

    def check_record_pack_unpack(self):
        oid = '0'*8
        data = 'sample'
        reflist = ['1'*8, '2'*8]
        refs = ''.join(reflist)
        self.test_stmt("result=unpack_record(pack_record(oid, data, refs))")
        self.test_val("result[0]", oid)
        self.test_val("result[1]", data)
        self.test_val("split_oids(result[2])", reflist)
        self.test_val("split_oids('')", [])

if __name__ == "__main__":
    (scenarios, options) = parse_args()
    options.show_coverage = options.show_coverage_lines = 1
    run_scenarios(scenarios, options)



#!/www/python/bin/python
"""
$URL: svn+ssh://svn/repos/trunk/durus/test/test_persistent.py $
$Id: test_persistent.py 23909 2004-04-07 13:37:02Z dbinger $
"""
from sancho.unittest import TestScenario, parse_args, run_scenarios
from durus.persistent import Persistent
from durus.file_storage import TempFileStorage
from durus.connection import Connection
from durus.utils import p64

tested_modules = ["durus.persistent"]

class Test (TestScenario):

    def check_getstate(self):
        self.test_stmt("p=Persistent()")
        self.test_val("p.__getstate__()", {})
        self.test_stmt("p.a = 1")
        self.test_val("p.__getstate__()", {'a':1})

    def check_setstate(self):
        self.test_stmt("p=Persistent()")
        self.test_stmt("p.__setstate__({})")
        self.test_stmt("p.__setstate__({'a':1})")
        self.test_val("p.a", 1)

    def check_change(self):
        self.test_stmt("p=Persistent()")
        self.test_stmt("p._p_changed == 0")
        self.test_stmt("p._p_note_change()")
        self.test_val("p._p_changed", True)

    def check_accessors(self):
        self.test_stmt("p=Persistent()")
        self.test_stmt("p._p_oid")
        self.test_val("p._p_format_oid()", 'None')
        self.test_stmt("p._p_oid = p64(1)")
        self.test_val("p._p_format_oid()", '1')
        self.test_val("repr(p)", "<Persistent 1>")

    def check_more(self):
        storage = TempFileStorage()
        connection = Connection(storage)
        self.test_stmt("root=connection.get_root()")
        self.test_true("root._p_is_ghost()")
        self.test_stmt("root.a = 1")
        self.test_true("root._p_is_unsaved()")
        self.test_stmt("del root.a")
        self.test_stmt("connection.abort()")
        self.test_true("root._p_is_ghost()")
        self.test_exc("root.a", AttributeError)
        self.test_stmt("root._p_set_status_saved()")
        self.test_true("root._p_is_saved()")
        self.test_stmt("root._p_set_status_unsaved()")
        self.test_true("root._p_is_unsaved()")
        self.test_stmt("root._p_set_status_ghost()")
        self.test_true("root._p_is_ghost()")
        self.test_stmt("root._p_set_status_unsaved()")

if __name__ == "__main__":
    (scenarios, options) = parse_args()
    options.show_coverage = options.show_coverage_lines = 1
    run_scenarios(scenarios, options)



#!/www/python/bin/python
"""$URL: svn+ssh://svn/repos/trunk/durus/storage_server.py $
$Id: storage_server.py 24846 2004-08-01 01:42:50Z dbinger $
"""
import socket
import select
import errno
from sets import Set
from durus.logger import log, is_logging
from durus.utils import p32, u32, u64
from durus.serialize import extract_class_name


STATUS_OKAY = 'O'
STATUS_KEYERROR = 'K'

TIMEOUT = 10
DEFAULT_HOST = '127.0.0.1'
DEFAULT_PORT = 2972


def recv(s, n):
    """(s:socket, n:int) -> str
    Call the recv() method on the socket, repeating as required until n bytes
    are received.  
    """
    data = []
    while n > 0:
        hunk = s.recv(n)
        if not hunk:
            raise IOError, 'connection reset by peer'
        n -= len(hunk)
        data.append(hunk)
    return ''.join(data)

class _Client:

    def __init__(self, s, addr):
        self.s = s
        self.addr = addr
        self.invalid = Set()

class ClientError(Exception):
    pass

class StorageServer:

    def __init__(self, storage, host=DEFAULT_HOST, port=DEFAULT_PORT):
        self.storage = storage
        self.clients = []
        self.sockets = []
        self.host = host
        self.port = port

    def serve(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            sock.bind((self.host, self.port))
        except socket.error, exc:
            if exc.args[0] == 98:
                raise RuntimeError(
                    "Port %s on %s is already in use by another process." %
                    (self.port, self.host))
        sock.listen(40)
        log(20, 'Ready with %s objects', self.storage.get_size())
        self.sockets.append(sock)
        while 1:
            r, w, e = select.select(self.sockets, [], [])
            for s in r:
                if s is sock:
                    # new connection
                    conn, addr = s.accept()
                    conn.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
                    conn.settimeout(TIMEOUT)
                    self.clients.append(_Client(conn, addr))
                    self.sockets.append(conn)
                else:
                    # command from client
                    try:
                        self.handle(s)
                    except (ClientError, socket.error, socket.timeout), exc:
                        log(10, '%s', ''.join(map(str, exc.args)))
                        self.sockets.remove(s)
                        self.clients.remove(self._find_client(s))

    def handle(self, s):
        command_code = s.recv(1)
        if not command_code:
            raise ClientError('EOF from client')
        handler = getattr(self, 'handle_%s' % command_code)
        handler(s)

    def _find_client(self, s):
        for client in self.clients:
            if client.s is s:
                return client
        assert 0

    def handle_N(self, s):
        # new OID
        s.sendall(self.storage.new_oid())

    def handle_L(self, s):
        # load
        oid = recv(s, 8)
        try:
            record = self.storage.load(oid)
        except KeyError:
            log(10, 'KeyError %s', u64(oid))
            s.sendall(STATUS_KEYERROR)
        else:
            if is_logging(5):
                log(5, 'Load %-7s %s', u64(oid), extract_class_name(record))
            s.sendall(STATUS_OKAY + p32(len(record)) + record)

    def handle_C(self, s):
        # commit
        client = self._find_client(s)
        s.sendall(p32(len(client.invalid)) + ''.join(client.invalid))
        client.invalid.clear()
        tlen = u32(recv(s, 4))
        if tlen == 0:
            return # client decided not to commit (e.g. conflict)
        tdata = recv(s, tlen)
        logging_debug = is_logging(10)
        logging_debug and log(10, 'Committing %s bytes', tlen)
        self.storage.begin()
        i = 0
        oids = []
        while i < len(tdata):
            rlen = u32(tdata[i:i+4])
            i += 4
            record = tdata[i:i+rlen]
            i += rlen
            oid = record[:8]
            if logging_debug:
                # extract_class_name() expects a record with a tid.
                class_name = extract_class_name('faketid:' + record)
                log(10, '  oid=%-6s rlen=%-6s %s', u64(oid), rlen, class_name)
            # storage will add tid
            self.storage.store(record)
            oids.append(oid)
        assert i == len(tdata)
        tid = self.storage.end()
        log(20, 'Committed %3s objects %s bytes', len(oids), tlen)
        s.sendall(STATUS_OKAY + tid)
        for c in self.clients:
            if c is not client:
                c.invalid.update(oids)

    def handle_S(self, s):
        # sync
        client = self._find_client(s)
        log(10, 'Sync %s', len(client.invalid))
        tid, invalid = self.storage.sync()
        assert not invalid # should have exclusive access
        s.sendall(tid + p32(len(client.invalid)) + ''.join(client.invalid))
        client.invalid.clear()

    def handle_P(self, s):
        # pack
        log(20, 'Pack')
        self.storage.pack()
        s.sendall(STATUS_OKAY)

    def handle_Q(self, s):
        # graceful quit
        log(20, 'Quit')
        raise SystemExit


def wait_for_server(host, port, maxtries=30, sleeptime=2):
    # Wait for the server to bind to the port.
    import time
    for attempt in range(maxtries):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            sock.connect((host, port))
        except socket.error, e:
            if not e.args[0] == errno.ECONNREFUSED:
                raise
            time.sleep(sleeptime)
        else:
            break
    else:
        raise SystemExit('Timeout waiting for port.')
    sock.close()



"""$URL: svn+ssh://svn/repos/trunk/durus/persistent_dict.py $
$Id: persistent_dict.py 23964 2004-04-12 20:39:19Z dbinger $
"""

from copy import copy
from durus.persistent import Persistent

class PersistentDict(Persistent):

    """
    Instance attributes:
      data : dict
    """
    data_is = dict

    def __init__(self, *args, **kwargs):
        self.data = dict(*args, **kwargs)

    def __cmp__(self, dict):
        if isinstance(dict, PersistentDict):
            return cmp(self.data, dict.data)
        else:
            return cmp(self.data, dict)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, item):
        self._p_note_change()
        self.data[key] = item

    def __delitem__(self, key):
        self._p_note_change()
        del self.data[key]

    def clear(self):
        self._p_note_change()
        self.data.clear()

    def copy(self):
        if self.__class__ is PersistentDict:
            return PersistentDict(self.data)
        # Use the copy module to copy self without data, and then use the
        # update method to fill the data in the new instance.
        changed = self.get_p_changed()
        data = self.data
        try:
            self.data = {} # This is why we saved _p_changed.
            c = copy(self)
        finally:
            self.data = data
            self._p_note_change(changed)
        c.update(self)
        return c

    def keys(self):
        return self.data.keys()

    def items(self):
        return self.data.items()

    def iteritems(self):
        return self.data.iteritems()

    def iterkeys(self):
        return self.data.iterkeys()

    def itervalues(self):
        return self.data.itervalues()

    def values(self):
        return self.data.values()

    def has_key(self, key):
        return self.data.has_key(key)

    def update(self, other):
        self._p_note_change()
        if isinstance(other, PersistentDict):
            self.data.update(other.data)
        elif isinstance(other, dict):
            self.data.update(other)
        else:
            for k, v in dict.items():
                self[k] = v

    def get(self, key, failobj=None):
        return self.data.get(key, failobj)

    def setdefault(self, key, failobj=None):
        if key not in self.data:
            self._p_note_change()
            self.data[key] = failobj
            return failobj
        return self.data[key]

    def pop(self, key, *args):
        self._p_note_change()
        return self.data.pop(key, *args)

    def popitem(self):
        self._p_note_change()
        return self.data.popitem()

    def __contains__(self, key):
        return key in self.data

    def fromkeys(cls, iterable, value=None):
        d = cls()
        for key in iterable:
            d[key] = value
        return d
    fromkeys = classmethod(fromkeys)

    def __iter__(self):
        return iter(self.data)




#!/www/python/bin/python
"""
$URL: svn+ssh://svn/repos/trunk/durus/test/test_file_storage.py $
$Id: test_file_storage.py 23897 2004-04-06 20:56:28Z dbinger $
"""
from sets import Set
from sancho.unittest import TestScenario, parse_args, run_scenarios
from durus.file_storage import TempFileStorage
from durus.serialize import pack_record
from durus.utils import p64

tested_modules = ["durus.file_storage"]

class Test (TestScenario):

    def check_file_storage(self):
        self.test_stmt("b=TempFileStorage()")
        self.test_val("b._get_tid()", p64(0))
        self.test_val("b.new_oid()", p64(1))
        self.test_val("b.new_oid()", p64(2))
        self.test_exc("b.load(p64(0))", KeyError)
        record = pack_record(p64(0), 'ok', '')
        self.test_stmt("b.store(record)")
        self.test_stmt("b.begin()")
        self.test_val("b._get_tid()", p64(1))
        self.test_stmt("b.end()")
        self.test_stmt("b.sync()")
        self.test_stmt("b.begin()")
        self.test_stmt("b.store(pack_record(p64(1), 'no', ''))")
        self.test_stmt("b.end()")
        self.test_stmt("b.pack()")


if __name__ == "__main__":
    (scenarios, options) = parse_args()
    options.show_coverage = options.show_coverage_lines = 1
    run_scenarios(scenarios, options)

